"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = require("axios");
const https_1 = require("https");
const log_1 = require("../logger/log");
const ts_md5_1 = require("ts-md5");
const apiError_1 = require("../errors/apiError");
const loggingDecorator_1 = require("../decorators/loggingDecorator");
const apiUtils_1 = require("./apiUtils");
const tls = require("tls");
const readline = require("readline");
const fs = require("fs");
const SID_HEADER = 'X-chkp-sid';
const CONTENT_LENGTH_HEADER = 'Content-Length';
const SID = 'sid';
const TASK_ID = 'task-id';
const SHOW_TASK = 'show-task';
const SENSITIVE_REQUEST_COMMANDS = ['login', 'add-administrator', 'set-administrator', 'add-opsec-application', 'set-opsec-application',
    'add-vpn-community-meshed', 'set-vpn-community-meshed', 'add-vpn-community-star', 'set-vpn-community-star', 'add-simple-gateway',
    'set-simple-gateway', 'add-data-center-server', 'set-data-center-server', 'delete-api-key', 'add-server-certificate',
    'set-server-certificate', 'add-user', 'set-user', 'add-mds', 'set-mds', 'add-checkpoint-host', 'set-checkpoint-host'];
const FINGERPRINTS_FILE = './fingerprints.json';
let headers = {
    'User-Agent': 'typescript-api-wrapper',
    'Accept': '*/*',
    'Content-Type': 'application/json'
};
class ApiClient {
    constructor({ port = 443, fingerprint, sid, server = '127.0.0.1', proxy, apiVersion, unsafe = false, unsafeAutoAccept = false, context = 'web_api' }) {
        this._port = port;
        this._fingerprint = fingerprint;
        this._sid = sid;
        this._server = server;
        this._proxy = proxy;
        this._apiVersion = apiVersion;
        this._unsafe = unsafe;
        this._unsafeAutoAccept = unsafeAutoAccept;
        this._approvedFingerprint = false;
        this._context = context;
        this._agent = new https_1.Agent({
            rejectUnauthorized: false,
            keepAlive: true
        });
    }
    login({ username, password, apiKey, continueLastSession = false, domain, readOnly = false }) {
        let credentials = {};
        if (username) {
            credentials['user'] = username;
        }
        if (password) {
            credentials['password'] = password;
        }
        if (apiKey) {
            credentials['api-key'] = apiKey;
        }
        if (this._context === 'web_api') {
            credentials['continue-last-session'] = continueLastSession;
            credentials['read-only'] = readOnly;
        }
        if (domain != null) {
            credentials['domain'] = domain;
        }
        return this.apiCall('login', credentials);
    }
    logout() {
        return this.apiCall('logout')
            .then(response => {
            this.agent.destroy();
            return response;
        });
    }
    publish() {
        return this.apiCall('publish');
    }
    discard() {
        return this.apiCall('discard');
    }
    async apiCall(command, payload = {}, sid, waitForTask = true) {
        function logRequest(config, command) {
            function isSensitiveRequest(command) {
                return SENSITIVE_REQUEST_COMMANDS.includes(command);
            }
            let headersForrLog = JSON.parse(JSON.stringify(config.headers));
            if (SID_HEADER in headersForrLog) {
                let hashedSid = ts_md5_1.Md5.hashStr(headersForrLog[SID_HEADER]);
                headersForrLog[SID_HEADER] = hashedSid;
            }
            let payloadForLog = config.data;
            if (isSensitiveRequest(command)) {
                payloadForLog = '****';
            }
            logger.info(`API REQUEST:
COMMAND: ${command}
HEADERS: ${JSON.stringify(headersForrLog, null, 2)}
PAYLOAD: ${JSON.stringify(payloadForLog, null, 2)}`);
        }
        if (!await this.isFingerprintOK()) {
            throw new apiError_1.default('Fingerprint is not OK.');
        }
        headers[CONTENT_LENGTH_HEADER] = Object.keys(payload).length; // If we run node.js need this header, and if we run through browser we don't
        if (sid != null) {
            this.sid = sid;
        }
        if (this.sid != null) {
            headers[SID_HEADER] = this.sid;
        }
        let url = `https://${this._server}:${this._port}/${this._context}/${(this._apiVersion ? `v${this._apiVersion}/` : "")}${command}`;
        const config = {
            method: 'post',
            url: url,
            headers: headers,
            data: payload,
            httpsAgent: this.agent
            // httpsAgent: new Agent({ // If there will be parallel problems, we can create new Agent (without keepAlive) for each API request.
            //     rejectUnauthorized: false
            // })
        };
        if (this.proxy != null) {
            config.proxy = this.proxy;
        }
        logRequest(config, command);
        return axios_1.default(config)
            .then(async (response) => {
            // handle success
            function logResponse(data, command) {
                function isSensitiveReply(command) {
                    return command === 'add-api-key';
                }
                if (!isSensitiveReply(command)) {
                    let dataForLog = data;
                    if (containsSid(command, data)) {
                        dataForLog = JSON.parse(JSON.stringify(data));
                        let hashedSid = ts_md5_1.Md5.hashStr(dataForLog[SID]);
                        dataForLog[SID] = hashedSid;
                    }
                    const responseMessage = `API RESPONSE:
COMMAND: ${command}
DATA: ${JSON.stringify(dataForLog, null, 2)}`;
                    logger.info(responseMessage);
                    console.log(responseMessage);
                }
            }
            function containsSid(command, data) {
                return (command === 'login' || command === 'login-to-domain') && SID in data;
            }
            logResponse(response.data, command);
            if (containsSid(command, response.data)) {
                this.sid = response.data.sid;
            }
            let responseToUser = response;
            if (waitForTask && command != SHOW_TASK && command != 'show-tasks') {
                if (TASK_ID in response.data) {
                    responseToUser = await this.waitForTaskFun(response.data[TASK_ID]);
                }
                else if ('tasks' in response.data) {
                    responseToUser = await this.waitForTasksFun(response.data['tasks']);
                }
            }
            return responseToUser;
        })
            .catch(error => {
            throw new apiError_1.default(error);
        });
    }
    async isFingerprintOK() {
        if (this.unsafe || this.approvedFingerprint) {
            return true;
        }
        const serverFingerprint = await this.getServerFingerprint();
        if (serverFingerprint === '') {
            throw new apiError_1.default('Fingerprint returned from the server is empty.');
        }
        if (this.fingerprint === serverFingerprint) {
            logger.info('Server Fingerprint equals to field fingerprint');
            this.approvedFingerprint = true;
            return true;
        }
        const localFingerprint = this.readFingerprintFromFile();
        if (serverFingerprint === localFingerprint) {
            logger.info('Server Fingerprint equals to local fingerprint');
            this.fingerprint = serverFingerprint;
            this.approvedFingerprint = true;
            return true;
        }
        if (!this.unsafeAutoAccept) {
            if (localFingerprint === '') {
                console.log('You currently do not have a record of this server\'s fingerprint.');
            }
            else {
                console.log('The server\'s fingerprint is different from your local record of this server\'s fingerprint.\n' +
                    'You maybe a victim to a Man-in-the-Middle attack, please beware.');
            }
            console.log(`Server Fingerprint (SHA-256): ${serverFingerprint}`);
            if (!await this.askYesNoQuestion('Do you accept this fingerprint?')) {
                return false;
            }
        }
        this.saveFingerprintToFile(serverFingerprint);
        this.fingerprint = serverFingerprint;
        this.approvedFingerprint = true;
        return true;
    }
    askYesNoQuestion(question) {
        let rl = readline.createInterface({
            input: process.stdin,
            output: process.stdout
        });
        return new Promise(resolve => {
            rl.question(question + ' [y/n]\n', (answer) => {
                let lowerCaseAnswer = answer.toLowerCase();
                const userAnswer = lowerCaseAnswer === 'y' || lowerCaseAnswer === 'yes';
                logger.info(`User answer for whether he accepts the fingerprint is: ${userAnswer}`);
                resolve(userAnswer);
            });
        }).then(respone => {
            rl.close();
            return respone;
        });
    }
    saveFingerprintToFile(serverFingerprint) {
        let fingerprints = ApiClient.getAllFingerprints();
        fingerprints[this.server] = serverFingerprint;
        fs.promises.writeFile(FINGERPRINTS_FILE, JSON.stringify(fingerprints, null, 2), { flag: 'w' });
    }
    readFingerprintFromFile() {
        let fingerprints = ApiClient.getAllFingerprints();
        if (this.server in fingerprints) {
            return fingerprints[this.server];
        }
        return '';
    }
    static isFingerprintsFileExist() {
        return fs.existsSync(FINGERPRINTS_FILE);
    }
    static getAllFingerprints() {
        let fingerprints = {};
        if (ApiClient.isFingerprintsFileExist()) {
            try {
                let fingerprintsStr = fs.readFileSync(FINGERPRINTS_FILE).toString();
                fingerprints = JSON.parse(fingerprintsStr);
            }
            catch (e) {
                logger.error('The fingerprint file is not in Json format.', e);
            }
        }
        return fingerprints;
    }
    getServerFingerprint() {
        let fingerprint;
        let options = {
            host: this.server,
            port: this.port,
            rejectUnauthorized: false
        };
        return new Promise((resolve) => {
            let client = tls.connect(options, () => {
                fingerprint = client.getPeerCertificate(false).fingerprint256;
                logger.info(`Server Fingerprint (SHA-256): ${fingerprint}`);
                resolve(fingerprint);
                client.end(() => {
                    logger.info('Client closed successfully after getting the server fingerprint.');
                });
            });
        });
    }
    async waitForTasksFun(tasks) {
        let taskIds = [];
        tasks.forEach(task => {
            if (TASK_ID in task) {
                taskIds.push(task[TASK_ID]);
                this.waitForTaskFun(TASK_ID);
            }
        });
        return await this.apiCall(SHOW_TASK, { 'task-id': taskIds, 'details-level': 'full' });
    }
    async waitForTaskFun(taskId) {
        let taskComplete = false;
        let taskResult;
        do {
            taskResult = await this.apiCallFewTimesOnError(SHOW_TASK, { 'task-id': taskId, 'details-level': 'full' }, 5);
            let commpletedTasks = taskResult.data['tasks'].filter((task) => task['status'] != 'in progress').length;
            let totalTasks = taskResult.data['tasks'].length;
            if (commpletedTasks == totalTasks) {
                taskComplete = true;
            }
            else {
                await apiUtils_1.default.sleep(2000);
            }
        } while (!taskComplete);
        return taskResult;
    }
    async apiCallFewTimesOnError(command, payload, attempts) {
        return this.apiCall(command, payload)
            .catch(error => {
            if (attempts <= 1) {
                throw new apiError_1.default(error);
            }
            return this.apiCallFewTimesOnError(command, payload, attempts - 1);
        });
    }
    get port() {
        return this._port;
    }
    set port(value) {
        this._port = value;
    }
    get fingerprint() {
        return this._fingerprint;
    }
    set fingerprint(value) {
        this._fingerprint = value;
    }
    get sid() {
        return this._sid;
    }
    set sid(value) {
        this._sid = value;
    }
    get server() {
        return this._server;
    }
    set server(value) {
        if (this.server != value) {
            this.approvedFingerprint = false;
        }
        this._server = value;
    }
    get proxy() {
        return this._proxy;
    }
    set proxy(value) {
        this._proxy = value;
    }
    get apiVersion() {
        return this._apiVersion;
    }
    set apiVersion(value) {
        this._apiVersion = value;
    }
    get unsafe() {
        return this._unsafe;
    }
    set unsafe(value) {
        this._unsafe = value;
    }
    get unsafeAutoAccept() {
        return this._unsafeAutoAccept;
    }
    set unsafeAutoAccept(value) {
        this._unsafeAutoAccept = value;
    }
    get approvedFingerprint() {
        return this._approvedFingerprint;
    }
    set approvedFingerprint(value) {
        this._approvedFingerprint = value;
    }
    get context() {
        return this._context;
    }
    set context(value) {
        this._context = value;
    }
    get agent() {
        return this._agent;
    }
    set agent(value) {
        this._agent = value;
    }
}
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "logout", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "publish", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "discard", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "apiCall", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "isFingerprintOK", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "askYesNoQuestion", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "saveFingerprintToFile", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "readFingerprintFromFile", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "getServerFingerprint", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "waitForTasksFun", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "waitForTaskFun", null);
__decorate([
    loggingDecorator_1.default()
], ApiClient.prototype, "apiCallFewTimesOnError", null);
exports.default = ApiClient;
const logger = log_1.default(ApiClient.name);
//# sourceMappingURL=apiClient.js.map